import mimetypes
import os
from tempfile import SpooledTemporaryFile

import magic
import requests
from django.core.files import File
from django.core.files.storage import FileSystemStorage
from django.utils.deconstruct import deconstructible
from django.utils.functional import LazyObject

from .utils import get_default_params


@deconstructible
class UrlStorage(FileSystemStorage):

    def get_available_name(self, name, *args, **kwargs):
        return name

    def _open(self, url, mode='rb'):
        params = get_default_params()
        params['url'] = url
        file = SpooledTemporaryFile(mode='b')
        file.write(requests.get(**params).content)
        file.seek(0)
        return File(file, mode)

    def _save(self, url, content):
        params = get_default_params()
        params['url'] = url
        params.setdefault('headers', {})['Content-Type'] = magic.from_buffer(content.read(100), mime=True)
        content.seek(0)
        response = requests.post(data=content.read(), **params)
        response.raise_for_status()
        return url

    def exists(self, name):
        return True

    def listdir(self, path):
        return [], []

    def url(self, name):
        return name


class DefaultStorage(LazyObject):
    def _setup(self):
        self._wrapped = UrlStorage()


default_storage = DefaultStorage()



import logging
import requests
from django.conf import settings
from django.db.models.query import ModelIterable, QuerySet
from django.db.models.sql import Query
from django.db.models.sql import Query
from requests import HTTPError
from rest_framework.exceptions import ValidationError
from urllib.parse import urlencode

from django_urlqueryset.utils import get_default_params

logger = logging.getLogger(__name__)


class UrlModelIterable(ModelIterable):
    def __iter__(self):
        return self.queryset.deserialize(self.queryset._result_cache)


class UrlQuery(Query):
    def __init__(self, model, **kwargs):
        self.model = model
        self.filters = {}
        self.order_by = []
        self.high_mark = settings.URLQS_HIGH_MARK
        self.low_mark = 0
        self.nothing = False
        self.distinct_fields = []
        super(UrlQuery, self).__init__(model, **kwargs)

    def get_meta(self):
        return self.model._meta

    def set_limits(self, low=None, high=None):
        if high is not None:
            if self.high_mark is not None:
                self.high_mark = min(self.high_mark, self.low_mark + high)
            else:
                self.high_mark = self.low_mark + high
        if low is not None:
            if self.high_mark is not None:
                self.low_mark = min(self.high_mark, self.low_mark + low)
            else:
                self.low_mark = self.low_mark + low

        if self.low_mark == self.high_mark:
            self.set_empty()

    def set_empty(self):
        self.nothing = True

    def clear_ordering(self, force_empty):
        self.order_by = []

    def add_ordering(self, *ordering):
        if ordering:
            self.order_by += ordering

    def clone(self):
        new = self.__class__(self.model)
        new.filters = self.filters.copy()
        new.high_mark = self.high_mark
        new.low_mark = self.low_mark
        new.nothing = self.nothing
        new.order_by = self.order_by
        return new

    def can_filter(self):
        return True

    def chain(self):
        return self.clone()

    def add_q(self, q_object):
        self.filters.update(dict(q_object.children))

    def _execute(self, request_params, method='get', **kwargs):
        query_params = {}
        if self.high_mark is not None:
            query_params['offset'] = self.low_mark
        if self.low_mark is not None:
            query_params['limit'] = self.high_mark - self.low_mark
        if self.order_by:
            query_params['ordering'] = ','.join(self.order_by)
        elif self.get_meta().ordering:
            query_params['ordering'] = ','.join(self.get_meta().ordering)
        query_params.update(self.filters)
        for key, value in query_params.items():
            if key.endswith('__in') and isinstance(value, (list, tuple)):
                query_params[key] = ",".join(str(i) for i in value)
        _request_params = get_default_params()
        _request_params.update(request_params.copy())
        _request_params.update(kwargs)
        url = _request_params.pop('url').replace('{{model._meta.model_name}}', self.model._meta.model_name)
        if query_params:
            url = f"{url}?{urlencode(query_params, safe=',')}"
        response = getattr(requests, method)(url=url, **_request_params)
        response.raise_for_status()
        return response.json() if response.headers.get('Content-Type') == 'application/json' else response


class UrlQuerySet(QuerySet):
    def __init__(self, *args, **kwargs):
        self.request_params = kwargs.pop('request_params', {})
        super().__init__(*args, **kwargs)
        if isinstance(self.query, Query):
            self.query = UrlQuery(self.model)
        self._iterable_class = UrlModelIterable
        self._count = None
        self._result_cache = None

    def _clone(self):
        c = super()._clone()
        c.request_params = self.request_params
        return c

    def as_manager(cls, **request_params):
        from django.db.models.manager import Manager

        class _Manager(Manager.from_queryset(cls)):
            def get_queryset(self):
                queryset = super().get_queryset()
                queryset.request_params = self.request_params
                return queryset
        manager = _Manager()
        manager._built_with_as_manager = True
        manager.request_params = request_params
        return manager
    as_manager.queryset_only = True
    as_manager = classmethod(as_manager)

    def count(self):
        if self._count is None:
            qs = self._chain()
            qs.query.set_limits(0, 1)
            return qs.query._execute(self.request_params)[settings.URLQS_COUNT]
        return self._count

    def _fetch_all(self):
        if self._count is None:
            response = self.query._execute(self.request_params)
            self._result_cache = list(self.deserialize(response[settings.URLQS_RESULTS]))
            self._count = response[settings.URLQS_COUNT]

    def create(self, **kwargs):
        try:
            response = self.query._execute(self.request_params, method='post', json=kwargs)
            return list(self.deserialize([response]))[0]
        except HTTPError as e:
            raise ValidationError({'remote_api_error': e.response.json()})

    def delete(self, **kwargs):
        try:
            response = self.query._execute(self.request_params, method='delete', json=kwargs)
            return response
        except HTTPError as e:
            raise ValidationError({'remote_api_error': e.response.json()})

    def update(self, **kwargs):
        return self._chain().query._execute(self.request_params, method='patch', json=kwargs)

    def deserialize(self, json_data=()):
        related_fields = {}
        for field in self.model._meta.fields:
            if field.related_model:
                related_fields[field.name] = {'manager': field.related_model.objects, 'pks': []}

        for obj_data in json_data:
            for field_name, value in related_fields.items():
                if field_name in obj_data and obj_data[field_name]:
                    related_fields[field_name]['pks'].append(obj_data[field_name])

        for rel_field, data in related_fields.items():
            if data['pks']:
                data['objs'] = {obj.pk: obj for obj in data['manager'].filter(pk__in=data['pks'])}

        for obj_data in json_data:
            obj = self.model()
            for field, value in obj_data.items():
                # if isinstance(value, list):
                #     value = next(iter(value), '')
                if value and field in related_fields.keys():
                    setattr(obj, field, related_fields[field]['objs'][value])
                else:
                    setattr(obj, field, value)
            yield obj

    def __getitem__(self, k):
        """Retrieve an item or slice from the set of results."""
        if not isinstance(k, (int, slice)):
            raise TypeError
        assert ((not isinstance(k, slice) and (k >= 0)) or
                (isinstance(k, slice) and (k.start is None or k.start >= 0) and
                 (k.stop is None or k.stop >= 0))), \
            "Negative indexing is not supported."

        # if self._result_cache is not None:
        #     return self._result_cache[k]

        if isinstance(k, slice):
            qs = self._chain()
            if k.start is not None:
                start = int(k.start)
            else:
                start = None
            if k.stop is not None:
                stop = int(k.stop)
            else:
                stop = None
            qs.query.set_limits(start, stop)
            return list(qs)[::k.step] if k.step else qs

        qs = self._chain()
        qs.query.set_limits(k, k + 1)
        qs._fetch_all()
        return qs._result_cache[0]





import logging
from django.apps import AppConfig
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import ugettext_lazy as _

logger = logging.getLogger('django_urlqueryset')


class CheckConfig(AppConfig):
    name = 'django_urlqueryset'

    def ready(self):
        if not self.check_settings():
            raise ImproperlyConfigured(_('django_urlqueryset is not '
                                         'configured correctly. '
                                         'Please check your settings.'))
        self.set_default_settings()

    def check_settings(self):

        return all([
            hasattr(settings, 'URLQS_HIGH_MARK'),
            hasattr(settings, 'URLQS_COUNT'),
            hasattr(settings, 'URLQS_RESULTS'),
        ])

    def set_default_settings(self):
        if not hasattr(settings, 'URLQS_HIGH_MARK'):
            setattr(settings, 'URLQS_HIGH_MARK', 1000000)
        if not hasattr(settings, 'URLQS_COUNT'):
            setattr(settings, 'URLQS_COUNT', 'count')
        if not hasattr(settings, 'URLQS_RESULTS'):
            setattr(settings, 'URLQS_RESULTS', 'results')





import pkg_resources
from .urlqueryset import UrlQuerySet
from .fields import UrlFileField, UrlImageField

__all__ = ('UrlQuerySet', 'UrlFileField', 'UrlImageField', 'VERSION')

__version__ = pkg_resources.get_distribution(__name__).version

VERSION = __version__.split('.')

default_app_config = 'django_urlqueryset.apps.CheckConfig'




from django.db import models
from  django.db.models.fields import files as file_fields

from .storages import default_storage
from .utils import get_default_params


class FieldFileMixin:
    def save(self, name, content, save=False):
        return super().save(name, content, save)


class FileFieldMixin:
    def generate_filename(self, instance, filename):
        url = getattr(instance, 'URL', None)
        if not url:
            params = get_default_params()
            url = params["url"].replace("{{model._meta.model_name}}", instance._meta.model_name)
        return f'{url}{instance.pk}/{self.name}/'


class UrlFieldFile(FieldFileMixin, file_fields.FieldFile):
    pass


class UrlImageFieldFile(FieldFileMixin, file_fields.ImageFieldFile):
    pass


class UrlFileField(FileFieldMixin, models.FileField):
    attr_class = UrlFieldFile

    def __init__(self, verbose_name=None, name=None, upload_to='', storage=None, **kwargs):
        storage = storage or default_storage
        super().__init__(verbose_name=verbose_name, name=name, upload_to=upload_to, storage=storage, **kwargs)


class UrlImageField(FileFieldMixin, models.ImageField):
    attr_class = UrlImageFieldFile

    def __init__(self, verbose_name=None, name=None, width_field=None, height_field=None, **kwargs):
        kwargs['storage'] = kwargs.get('storage') or default_storage
        super().__init__(verbose_name=verbose_name, name=name, width_field=width_field, height_field=height_field, **kwargs)


import os
from setuptools import find_packages, setup

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django_urlqueryset',
    version='0.4.6',
    packages=find_packages(),
    include_package_data=True,
    url='https://github.com/MaurizioPilia/django-urlqueryset',
    install_requires=[
        'Django>=1.8',
        'djangorestframework>=3.9.4',
        'requests>=2.11.1',
        'python-magic>=0.4.18'
    ],
    extras_require={
        'test': [
            'ipython',
            'ipdb',
            'Sphinx',
            'sphinx-rtd-theme',
        ]
    },
    zip_safe=False,
)
